#!/usr/bin/python2.2

import sys, os
import Image, ImageDraw # PIL
from math import *

# defaults
SCALE = 5000 # 1:5000
RESOL = 254/.0254 # 254dpi -> 10000dot/meter

[MIN, MAX] = range(2)

rad2deg = lambda r: r/pi*180
deg2rad = lambda d: d/180*pi
def dms2deg(d, m=0, s=0): return d + m/60.0 + s/3600.0

def WGSlalo_to_KKJlalo((latR, lonR)):
# Input:  latR  is Latitude in Radians, north is positive
#         lonR  is Longitude in Radians, east is positive
# Output: [0]  is Latitude in Radians, north is positive
#         [1]  is Longitude in Radians, east is positive
#
# Accuracy has some limitations in range of 0.2-1.0 meters
#
  lat, lon = rad2deg(latR), rad2deg(lonR)
  dLa = deg2rad( -0.124766E+01 +
                  0.269941E+00 * lat +
                 -0.191342E+00 * lon +
                 -0.356086E-02 * lat * lat +
                  0.122353E-02 * lat * lon +
                  0.335456E-03 * lon * lon ) / 3600.0
  dLo = deg2rad(  0.286008E+02       +
                 -0.114139E+01 * lat +
                  0.581329E+00 * lon +
                  0.152376E-01 * lat * lat +
                 -0.118166E-01 * lat * lon +
                 -0.826201E-03 * lon * lon ) / 3600.0
  return (latR + dLa, lonR + dLo)

def KKJLaLo_to_KKJxy((latR, lonR), ZoneNumber, Long0R):
# Inputs:  lonR = Longitude (Radians)
#          latR = Latitude  (Radians)
# Outputs: ED50[X], ED50[Y]  (meters)
#
# Following method is NOT excellent for more
# than about 2deg from midline.
# It might be usable within a meter for wider
# projections

  Lo = lonR - Long0R

  a, f  = 6378388.0, 1/297.0            # Hayford ellipsoid
  b  = (1.0 - f) * a
  bb = b*b               # b
  c  = (a / b) * a       # c   = a/b
  ee = (a * a - bb) / bb # e' = (a - b)/b
  n  = (a - b)/(a + b)   # n   = (a-b)/(a+b)
  nn = n*n               # n

  cosLa = cos(latR)
  NN = ee * cosLa * cosLa # Eta = e' cos Latitude
  LaF = atan(tan(latR) / cos(Lo * sqrt(1 + NN)))
  cosLaF = cos(LaF)
  t   = (tan(Lo) * cosLaF) / sqrt(1 + ee * cosLaF * cosLaF)
  A   = a / (1 + n)
  A1  = A * (1 + nn/4 + nn * nn / 64)
  A2  = A * 1.5 * n * (1 - nn / 8)
  A3  = A * 0.9375 * nn * (1 - nn / 4)
  A4  = A * 35/48.0 * nn * n
  X   = A1 * LaF - \
        A2 * sin(2*LaF) + \
        A3 * sin(4*LaF) - \
        A4 * sin(6*LaF)
  Y   = c * log(t + sqrt(1 + t*t)) + \
        500000.0 + ZoneNumber * 1000000.0
  return (X, Y)


def WGS2KKJ((lat, lon), kaista):

    latR = deg2rad(lat)
    lonR = deg2rad(lon)

    latKKJ, lonKKJ = WGSlalo_to_KKJlalo((latR, lonR))
    Long0R = deg2rad(kaista * 3.0 + 18.0)
    return KKJLaLo_to_KKJxy((latKKJ, lonKKJ), kaista, Long0R);

kkj2img = lambda a: int(a/SCALE*RESOL)

class MagellanUnit:

  def __init__(self, fn):
    self.fh = open(fn)
    self.reading = ''

  def read_trk(self):
    """returns (lat, lon) or None (in degreas)"""
    line = self.fh.readline()
    while line:
      if line.startswith('$PMGNTRK'):
        cmd,lat,N,lon,E,rest = line.split(',',5)
        assert N == 'N' and E == 'E'
        lat, lon = map(float, [lat, lon])
        lat = dms2deg(int(lat/100), lat%100, 0)
        lon = dms2deg(int(lon/100), lon%100, 0)
        return (lat, lon)
      line = self.fh.readline()

  def read_wpt(self):
    pass

def minmax(range, val):
  """Find min/max"""
  return min(range[MIN], val) or val, max(range[MAX], val)

trks = []
wpts = []
sumlon = count = 0

reader = MagellanUnit(sys.argv[1])
outfile = sys.argv[2]

assert os.path.splitext(outfile)[1].lower() == '.bmp', "Only BMP images supported"

# Read track
trk = reader.read_trk()
while trk:
  trks.append(trk)
  sumlon += trk[1]; count += 1
  trk = reader.read_trk()

meridian = int(((sumlon/count)-16.5)/3)
assert 0 <= meridian <= 5, 'KKJ - kaista %d (0 <= kaista <= 5)' % meridian

# Convert WGS --> KKJ and find min/max
Nrange = Irange = (0,0)
for i in range(len(trks)):
  N,I = trks[i] = WGS2KKJ(trks[i], meridian)
  Nrange = minmax(Nrange, N)
  Irange = minmax(Irange, I)

# Read waypoints and accept wpts inside and near (+- 0.5) the track box
Nrange2 = Nrange[MIN]/2 - Nrange[MAX]/2, Nrange[MIN]/2 + Nrange[MAX]/2*3
Irange2 = Irange[MIN]/2 - Irange[MAX]/2, Irange[MIN]/2 + Irange[MAX]/2*3
wpt = reader.read_wpt()
while wpt:
  N,I = WGS2KKJ(wpt, meridian)
  if Nrange[MIN] <= N <= Nrange2[MAX] and Irange[MIN] <= I <= Irange[MAX]:
    wpts.append((N,I))
    Nrange = minmax(Nrange, N)
    Irange = minmax(Irange, I)
  wpt = reader.read_wpt()

minX = kkj2img(Irange[MIN]); maxX = kkj2img(Irange[MAX])
minY = kkj2img(Nrange[MIN]); maxY = kkj2img(Nrange[MAX])

image = Image.new("RGB", (maxX-minX+1, maxY-minY+1), (255,255,255))
draw = ImageDraw.Draw(image)
x0 = y0 = None
for N, I in trks:
  x, y = kkj2img(I)-minX, maxY-kkj2img(N)
  #draw.point((x,y), fill=(255,0,0))
  draw.line([x-1,y,x+1,y], fill=(255,0,0))
  draw.line([x,y-1,x,y+1], fill=(255,0,0))
  if x0 != None:
    draw.line([x0,y0,x,y], fill=(128,0,0))
  x0, y0 = x, y

image.save(outfile)

# Patch resolution:
def o32(i): return chr(i&255) + chr(i>>8&255) + chr(i>>16&255) + chr(i>>24&255)
fh = open(outfile, "rb+"); fh.seek(38); fh.write(o32(int(RESOL)) + o32(int(RESOL))); fh.close()
